From: Boqun Feng <boqun.feng@gmail.com>
To: Ingo Molnar <mingo@kernel.org>, Peter Zijlstra <peterz@infradead.org>
Cc: "Will Deacon" <will@kernel.org>,
"Waiman Long" <longman@redhat.com>,
Miguel Ojeda <ojeda@kernel.org>,
Alice Ryhl <aliceryhl@google.com>,
linux-kernel@vger.kernel.org, rust-for-linux@vger.kernel.org,
Breno Leitao <leitao@debian.org>,
"Paul E. McKenney" <paulmck@kernel.org>,
Erik Lundgren <elundgren@meta.com>,
Eric Dumazet <edumazet@google.com>,
Boqun Feng <boqun.feng@gmail.com>
Subject: [PATCH 9/9] lockdep: Speed up lockdep_unregister_key() with expedited RCU synchronization
Date: Wed, 16 Jul 2025 07:48:18 -0700 [thread overview]
Message-ID: <20250716144818.47650-10-boqun.feng@gmail.com> (raw)
In-Reply-To: <20250716144818.47650-1-boqun.feng@gmail.com>
From: Breno Leitao <leitao@debian.org>
lockdep_unregister_key() is called from critical code paths, including
sections where rtnl_lock() is held. For example, when replacing a qdisc
in a network device, network egress traffic is disabled while
__qdisc_destroy() is called for every network queue.
If lockdep is enabled, __qdisc_destroy() calls lockdep_unregister_key(),
which gets blocked waiting for synchronize_rcu() to complete.
For example, a simple tc command to replace a qdisc could take 13
seconds:
# time /usr/sbin/tc qdisc replace dev eth0 root handle 0x1: mq
real 0m13.195s
user 0m0.001s
sys 0m2.746s
During this time, network egress is completely frozen while waiting for
RCU synchronization.
Use synchronize_rcu_expedited() instead to minimize the impact on
critical operations like network connectivity changes.
This improves 10x the function call to tc, when replacing the qdisc for
a network card.
# time /usr/sbin/tc qdisc replace dev eth0 root handle 0x1: mq
real 0m1.789s
user 0m0.000s
sys 0m1.613s
[boqun: Fixed the comment and add more information for the temporary
workaround, and add TODO information for hazptr]
Reported-by: Erik Lundgren <elundgren@meta.com>
Signed-off-by: Breno Leitao <leitao@debian.org>
Reviewed-by: Paul E. McKenney <paulmck@kernel.org>
Reviewed-by: Eric Dumazet <edumazet@google.com>
Signed-off-by: Boqun Feng <boqun.feng@gmail.com>
Link: https://lore.kernel.org/r/20250321-lockdep-v1-1-78b732d195fb@debian.org
---
kernel/locking/lockdep.c | 12 ++++++++++--
1 file changed, 10 insertions(+), 2 deletions(-)
diff --git a/kernel/locking/lockdep.c b/kernel/locking/lockdep.c
index 0c941418a215..2d4c5bab5af8 100644
--- a/kernel/locking/lockdep.c
+++ b/kernel/locking/lockdep.c
@@ -6616,8 +6616,16 @@ void lockdep_unregister_key(struct lock_class_key *key)
if (need_callback)
call_rcu(&delayed_free.rcu_head, free_zapped_rcu);
- /* Wait until is_dynamic_key() has finished accessing k->hash_entry. */
- synchronize_rcu();
+ /*
+ * Wait until is_dynamic_key() has finished accessing k->hash_entry.
+ *
+ * Some operations like __qdisc_destroy() will call this in a debug
+ * kernel, and the network traffic is disabled while waiting, hence
+ * the delay of the wait matters in debugging cases. Currently use a
+ * synchronize_rcu_expedited() to speed up the wait at the cost of
+ * system IPIs. TODO: Replace RCU with hazptr for this.
+ */
+ synchronize_rcu_expedited();
}
EXPORT_SYMBOL_GPL(lockdep_unregister_key);
--
2.39.5 (Apple Git-154)
prev parent reply other threads:[~2025-07-16 14:48 UTC|newest]
Thread overview: 10+ messages / expand[flat|nested] mbox.gz Atom feed top
2025-07-16 14:48 [GIT PULL] [PATCH 0/9] Locking changes for v6.17 Boqun Feng
2025-07-16 14:48 ` [PATCH 1/9] spi: spi-nxp-fspi: Check return value of devm_mutex_init() Boqun Feng
2025-07-16 14:48 ` [PATCH 2/9] leds: lp8860: " Boqun Feng
2025-07-16 14:48 ` [PATCH 3/9] locking/mutex: Mark devm_mutex_init() as __must_check Boqun Feng
2025-07-16 14:48 ` [PATCH 4/9] rust: sync: Add #[must_use] to Lock::try_lock() Boqun Feng
2025-07-16 14:48 ` [PATCH 5/9] locking/rwsem: Use OWNER_NONSPINNABLE directly instead of OWNER_SPINNABLE Boqun Feng
2025-07-16 14:48 ` [PATCH 6/9] locking/lockdep: Avoid struct return in lock_stats() Boqun Feng
2025-07-16 14:48 ` [PATCH 7/9] locking/lockdep: Change 'static const' variables to enum values Boqun Feng
2025-07-16 14:48 ` [PATCH 8/9] locking/mutex: Remove redundant #ifdefs Boqun Feng
2025-07-16 14:48 ` Boqun Feng [this message]
Reply instructions:
You may reply publicly to this message via plain-text email
using any one of the following methods:
* Save the following mbox file, import it into your mail client,
and reply-to-all from there: mbox
Avoid top-posting and favor interleaved quoting:
https://en.wikipedia.org/wiki/Posting_style#Interleaved_style
* Reply using the --to, --cc, and --in-reply-to
switches of git-send-email(1):
git send-email \
--in-reply-to=20250716144818.47650-10-boqun.feng@gmail.com \
--to=boqun.feng@gmail.com \
--cc=aliceryhl@google.com \
--cc=edumazet@google.com \
--cc=elundgren@meta.com \
--cc=leitao@debian.org \
--cc=linux-kernel@vger.kernel.org \
--cc=longman@redhat.com \
--cc=mingo@kernel.org \
--cc=ojeda@kernel.org \
--cc=paulmck@kernel.org \
--cc=peterz@infradead.org \
--cc=rust-for-linux@vger.kernel.org \
--cc=will@kernel.org \
/path/to/YOUR_REPLY
https://kernel.org/pub/software/scm/git/docs/git-send-email.html
* If your mail client supports setting the In-Reply-To header
via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line
before the message body.
This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox
all inboxes | Powered by JetHome®